#include "allegro.h"
#include <math.h>
#include <stdio.h>

#ifndef M_PI
#define M_PI		3.1415926535897932384626433832795f
#define M_HALFPI	1.5707963267948966192313216916398f
#define M_DOUBLEPI	6.283185307179586476925286766559f
#endif

#define SCR_W	800
#define SCR_H	600
#define SCR_BPP	8

#define RADS_TO_DEGS(v)	((v*180.0f) / M_PI)

struct Sector;

//#define PREBUILT_TEXTURES	466
struct TextureSt
{
	BITMAP *Image;
//	BITMAP *BitMapTab[PREBUILT_TEXTURES/2];
};

struct Wall
{
	fixed FirstVertex[2];
	fixed UnitTangent[2];
	unsigned int Angle;
	fixed Length;

	enum WallTypes
	{
		SOLID, PORTAL
	} Type;

	TextureSt *Texture;
	Sector *Next;
};

struct Sector
{
	Wall *WallArray;
	int NumWalls;
};

#define UNITS_PER_CIRCLE	8192
#define CIRCLE_MASK			(UNITS_PER_CIRCLE-1)
#define RIGHT_ANGLE			(UNITS_PER_CIRCLE >> 2)
//#define QUARTER_MASK		(RIGHT_ANGLE-1)
#define HALF_CIRCLE			(UNITS_PER_CIRCLE >> 1)

#define RADS_TO_UNITS(v)	((v*UNITS_PER_CIRCLE)/M_DOUBLEPI)

fixed *AntiDistortionTable;
fixed *ViewAngleTable;
fixed *CosineTable, *TanTable;
struct DispStruct
{
	TextureSt *Tex;
	fixed Height;
	int TexX;

	unsigned char FC;
} *DisplayTable;
unsigned char UFC = 0;

#define LookupCos(v)	CosineTable[v]
#define LookupSin(v)	LookupCos((RIGHT_ANGLE-v)&CIRCLE_MASK)
#define LookupTan(v)	TanTable[v&(HALF_CIRCLE-1)]

void AllocateTables(void)
{
	ViewAngleTable = new fixed[SCREEN_W];
	AntiDistortionTable = new fixed[SCREEN_W];

	DisplayTable = new DispStruct[SCREEN_W];

	int row = SCREEN_W;
	while(row--)
	{
		float AngleInRadians;

		AngleInRadians = M_PI - atan( 1.5f*(-0.5f + ((float)row / SCREEN_W)) );

		ViewAngleTable[row] = (AngleInRadians*UNITS_PER_CIRCLE) / M_DOUBLEPI;
		AntiDistortionTable[row] = ftofix( cos(AngleInRadians) );

		DisplayTable[row].FC = 0;
	}

	CosineTable = new fixed[UNITS_PER_CIRCLE];
	TanTable = new fixed[HALF_CIRCLE];

	int angle = UNITS_PER_CIRCLE;
	while(angle--)
	{
		float AngleInRadians;

		AngleInRadians = ((float)angle*M_DOUBLEPI) / (float)UNITS_PER_CIRCLE;
		CosineTable[angle] = ftofix(cos(AngleInRadians));
		if(angle < HALF_CIRCLE) TanTable[angle] = ftofix(tan(AngleInRadians));
	}
}

void FreeTables(void)
{
	delete[] AntiDistortionTable;
	delete[] ViewAngleTable;
	delete[] CosineTable;
	delete[] TanTable;
	delete[] DisplayTable;
}

bool UseVideoMemory;

void SetupTexture(TextureSt *tex, BITMAP *src)
{
/*	int c = PREBUILT_TEXTURES;
	while(c-=2)
	{
		int height = c;
		if(height > SCREEN_H) height = SCREEN_H;

		tex->BitMapTab[c>>1] = UseVideoMemory ? create_video_bitmap(src->w, height) : create_system_bitmap(src->w, height);
		if(!tex->BitMapTab[c>>1]) tex->BitMapTab[c>>1] = create_system_bitmap(src->w, height);

		stretch_blit(src, tex->BitMapTab[c>>1], 0, 0, src->w, src->h, 0, (height >> 1) - (c >> 1), src->w, c);
		bytes += src->w*height;
	}*/

	tex->Image = create_bitmap(src->w, src->h);
	blit(src, tex->Image, 0, 0, 0, 0, src->w, src->h);
}

void FreeTexture(TextureSt *tex)
{
	destroy_bitmap(tex->Image);
}

int WallCompFunc(const void *elem1, const void *elem2)
{
	if( (((Wall *)elem1)->Type == Wall::SOLID) && (((Wall *)elem2)->Type == Wall::SOLID)) return 0;
	if( (((Wall *)elem1)->Type == Wall::PORTAL) && (((Wall *)elem2)->Type == Wall::PORTAL)) return 0;
	if( (((Wall *)elem1)->Type == Wall::PORTAL) && (((Wall *)elem2)->Type == Wall::SOLID)) return -1;
	return 1;
}

void SetupSector(Sector *sect)
{
	/* sort wall list, really */
	qsort(sect->WallArray, sect->NumWalls, sizeof(Wall), WallCompFunc);
}

void SetupWall(Wall *w, TextureSt *texture, fixed *Vert1, fixed *Vert2, Sector *next)
{
	/* copy some fields */
	w->FirstVertex[0] = Vert1[0];
	w->FirstVertex[1] = Vert1[1];
	w->Texture = texture;

	/* all this would be preprocessed and stored in a file */
	w->UnitTangent[0] = Vert2[0] - Vert1[0];
	w->UnitTangent[1] = Vert2[1] - Vert1[1];

	w->Length = ftofix(
					sqrt(
						fixtof(w->UnitTangent[0])*fixtof(w->UnitTangent[0]) + 
						fixtof(w->UnitTangent[1])*fixtof(w->UnitTangent[1])
					));

	w->UnitTangent[0] = fixdiv(w->UnitTangent[0], w->Length);
	w->UnitTangent[1] = fixdiv(w->UnitTangent[1], w->Length);

	w->Angle = RADS_TO_UNITS( atan2(-w->UnitTangent[1], w->UnitTangent[0]) );

	w->Type = next ? Wall::PORTAL : Wall::SOLID;
	w->Next = next;
}

int LowLine, HighLine;

void BeginFrame(void)
{
	UFC++;
	LowLine = 0;
	HighLine = SCREEN_W-1;
}

int IVLow, IVHigh;
bool IsVisible(Wall *w, unsigned int viewangle, fixed x, fixed y, int Start, int End)
{
	/* quick fix! */
	fixed WallX, WallY, vecx, vecy;

	// get vector to wall
 	vecx = w->FirstVertex[0] - x;
	vecy = w->FirstVertex[1] - y;
		
	// get co-ordinates of origin in wall space
	WallY = -fixmul(vecx, w->UnitTangent[1])	+ fixmul(vecy, w->UnitTangent[0]);
	// check not a reverse face
	if(WallY > 0) return false;
	WallX = fixmul(vecx, w->UnitTangent[0]) + fixmul(vecy, w->UnitTangent[1]);

	// 'cast rays'
	if(LowLine > Start) Start = LowLine;
	if(HighLine < End) End = HighLine;
	IVLow = Start;
	IVHigh = -1;
	for(int r = Start; r <= End; r++)
		if(DisplayTable[r].FC != UFC)
		{
			// calculate angle in wall space
			unsigned int WallAngle;

			WallAngle = (viewangle - w->Angle + ViewAngleTable[r])&CIRCLE_MASK;

			if(((WallAngle+RIGHT_ANGLE-1)&CIRCLE_MASK) < ((RIGHT_ANGLE << 1)-2))//(WallAngle > UNITS_PER_CIRCLE*3/4) || (WallAngle < UNITS_PER_CIRCLE/4))
			{
				// get texture X
				fixed TexX;

				TexX = fixmul(WallY, LookupTan(WallAngle));

				if((TexX >= WallX) && (TexX-WallX) <= w->Length)
				{
					if(IVHigh < 0) IVLow = IVHigh = r;
					IVHigh = r;
				}
			}
		}

	return (IVHigh > 0);
}

void AddWall(Wall *w, unsigned int viewangle, fixed x, fixed y, int Start, int End)
{
	fixed WallX, WallY, vecx, vecy;

	// check if just plain facing the wrong way
	if(LowLine > Start) Start = LowLine;
	if(HighLine < End) End = HighLine;
	viewangle = (viewangle-w->Angle)&CIRCLE_MASK;

//	if(((viewangle+ViewAngleTable[Start]+HALF_CIRCLE)&CIRCLE_MASK) > RIGHT_ANGLE) return;
//	if(viewangle < -RIGHT_ANGLE) return;

	// get vector to wall
	vecx = w->FirstVertex[0] - x;
	vecy = w->FirstVertex[1] - y;
		
	// get co-ordinates of origin in wall space
	WallY = -fixmul(vecx, w->UnitTangent[1])	+ fixmul(vecy, w->UnitTangent[0]);
	// check not a reverse face
	if(WallY > 0) return;
	WallX = fixmul(vecx, w->UnitTangent[0]) + fixmul(vecy, w->UnitTangent[1]);

	// 'cast rays'
	for(int r = Start; r <= End; r++)
		if(DisplayTable[r].FC != UFC)
		{
			// calculate angle in wall space
			unsigned int WallAngle;

			WallAngle = (viewangle + ViewAngleTable[r])&CIRCLE_MASK;

			if(((WallAngle+RIGHT_ANGLE-1)&CIRCLE_MASK) < ((RIGHT_ANGLE << 1)-2))//(WallAngle > UNITS_PER_CIRCLE*3/4) || (WallAngle < UNITS_PER_CIRCLE/4))
			{
				// get texture X
				fixed TexX;

				TexX = fixmul(WallY, LookupTan(WallAngle));

				/*	TexX / Depth = sin(theta)
					WallY / Depth = cos(theta)

					=> TexX = WallY*tan(theta) = Depth*sin(theta) */

				if((TexX >= WallX) && (TexX-WallX) <= w->Length)
				{
					// convert to texture co-ordinates
					TexX = (TexX - WallX) >> 10;

					// 'draw'
					fixed Height;

					/*
					Depth = (WallY*AD[r]) / cos(W);
					Height = SCREEN_H / Depth

					=> Height = SCREEN_H*cos(W) / WallY*AD[r]
					*/
					Height = fixdiv( fixmul( itofix(SCREEN_H), LookupCos(WallAngle) ), fixmul(WallY, AntiDistortionTable[r]) );

					DisplayTable[r].FC = UFC;
					DisplayTable[r].Height = Height;
					DisplayTable[r].Tex = w->Texture;
					DisplayTable[r].TexX = TexX&63;

					if(r == LowLine) LowLine++;
				}
			}
		}

	while(DisplayTable[HighLine].FC == UFC) HighLine--;
}

void DrawSlither(BITMAP *dest, TextureSt *src, int destx, int desty, int desth, int sourcex)
{
	if(desth <= 0)
		return;

	unsigned int c, sy, syv;
	BITMAP *source = src->Image;

	sy = 0;
	syv = (64 << 20) / desth;

	if(desty < 0)
	{
		sy += -desty*syv;
		desth += desty;
		desty = 0;
	}

	if(desty + desth > SCREEN_H)
		desth = SCREEN_H-desty;

#if SCR_BPP == 8
	unsigned char *lptr = source->line[sourcex];
	c = desth;
	while(c--)
	{
		dest->line[c+desty][destx] = lptr[sy >> 20];
		sy += syv;
	}
#endif

#if (SCR_BPP == 16) || (SCR_BPP == 15)
	unsigned short *lptr = (unsigned short *)source->line[sourcex];
	c = desth;
	while(c--)
	{
		((unsigned short *)dest->line[c+desty])[destx] = lptr[sy >> 20];
		sy += syv;
	}
#endif

#if SCR_BPP == 32
	unsigned int *lptr = (unsigned int *)source->line[sourcex];
	c = desth;
	while(c--)
	{
		((unsigned int *)dest->line[c+desty])[destx] = lptr[sy >> 20];
		sy += syv;
	}
#endif

}

void EndFrame(BITMAP *target)
{
	int c = SCREEN_W;
	while(c--)
		if(DisplayTable[c].FC == UFC)
			DrawSlither(target, DisplayTable[c].Tex, c, (SCREEN_H>>1) - (DisplayTable[c].Height >> 17), DisplayTable[c].Height >> 16, DisplayTable[c].TexX);
		else
		{
			/* self heal small errors */
			if(c)
				DrawSlither(target, DisplayTable[c-1].Tex, c, (SCREEN_H>>1) - (DisplayTable[c-1].Height >> 17), DisplayTable[c-1].Height >> 16, DisplayTable[c-1].TexX);
			else
				DrawSlither(target, DisplayTable[c+1].Tex, c, (SCREEN_H>>1) - (DisplayTable[c+1].Height >> 17), DisplayTable[c+1].Height >> 16, DisplayTable[c+1].TexX);
		}
}

int sc;

void DrawSector(Sector *s, unsigned int viewangle, fixed x, fixed y, int start = 0, int end = 319)
{
	int c = s->NumWalls;

	/* draw solid walls first */
	while((c--) && s->WallArray[c].Type == Wall::SOLID)
		AddWall(&s->WallArray[c], viewangle, x, y, start, end);

	/* now consider portals */
	if(c >= 0)
	{
		do
		{
			if(IsVisible(&s->WallArray[c], viewangle, x, y, start, end))
			{
				DrawSector(s->WallArray[c].Next, viewangle, x, y, IVLow, IVHigh);
				sc++;
			}
		}
		while(c--);
	}
}

#define COLLISION_RADIUS	ftofix(0.3)

Sector *AttemptMotion(Sector *s, fixed *x, fixed *y, fixed addx, fixed addy)
{
	/* run around checking for collisions with walls */
	fixed Time = itofix(1), CollTime;
	Wall *Collider;

	while((Time > 16) && (addx || addy))
	{
		CollTime = itofix(1);
		Collider = NULL;

		int c = s->NumWalls;
		while(c--)
		{
			fixed OldWallY = -fixmul(s->WallArray[c].FirstVertex[0] - *x, s->WallArray[c].UnitTangent[1])	+ fixmul(s->WallArray[c].FirstVertex[1] - *y, s->WallArray[c].UnitTangent[0]);

			fixed vecx = s->WallArray[c].FirstVertex[0] - *x - addx;
			fixed vecy = s->WallArray[c].FirstVertex[1] - *y - addy;
			fixed WallY = -fixmul(vecx, s->WallArray[c].UnitTangent[1])	+ fixmul(vecy, s->WallArray[c].UnitTangent[0]);

			fixed WallX = fixmul(*x - s->WallArray[c].FirstVertex[0], s->WallArray[c].UnitTangent[0])	+ fixmul(*y - s->WallArray[c].FirstVertex[1], s->WallArray[c].UnitTangent[1]);

//			fixed difference = OldWallY ^ WallY;
//			difference &= ~15;

			if((abs(WallY-OldWallY) > 30) && (WallX > -COLLISION_RADIUS) && (WallX < s->WallArray[c].Length+COLLISION_RADIUS))
			{
				if(s->WallArray[c].Type == Wall::PORTAL)
				{
					if(WallY > 0)
					{
						fixed NewCollTime;
						NewCollTime = fixdiv(-OldWallY, WallY - OldWallY);

						if(NewCollTime < CollTime)
						{
							CollTime = NewCollTime;
							Collider = &s->WallArray[c];
						}
					}
				}
				else
				{
					WallY += COLLISION_RADIUS;
					OldWallY += COLLISION_RADIUS;

					if(WallY > 0)
					{
						fixed NewCollTime;
						NewCollTime = fixdiv(-OldWallY, WallY - OldWallY);

						if(NewCollTime < CollTime)
						{
							CollTime = NewCollTime;
							Collider = &s->WallArray[c];
						}
					}
				}
			}
		}

		if(Collider)
		{
			*x += fixmul(addx, CollTime);
			*y += fixmul(addy, CollTime);
			Time = fixmul(Time, 65536-CollTime);

			/* fix rounding errors */
/*			fixed WallY = -fixmul(*x, Collider->UnitTangent[1])	+ fixmul(*y, Collider->UnitTangent[0]);
			if(WallY > -COLLISION_RADIUS)
			{
				*x += fixmul(WallY + COLLISION_RADIUS, -Collider->UnitTangent[1]);
				*y += fixmul(WallY + COLLISION_RADIUS, Collider->UnitTangent[0]);
			}*/

			/* calculate remaining motion */
			addx = fixmul(addx, 65536-CollTime);
			addy = fixmul(addy, 65536-CollTime);

			if(Collider->Type == Wall::PORTAL)
			{
				/* quick-hack way of changing sector */
				return AttemptMotion(Collider->Next, x, y, addx, addy);
			}

			/* if not portal, rub against wall */
			fixed movement = fixmul(addy, Collider->UnitTangent[1]) + fixmul(addx, Collider->UnitTangent[0]);
			addx = fixmul(movement, Collider->UnitTangent[0]);
			addy = fixmul(movement, Collider->UnitTangent[1]);
		}
		else
			break;
	}

	*x += addx;
	*y += addy;
	return s;
}

volatile bool quit = false;

void CloseProc(void)
{
	quit = true;
}

int main(void)
{
	BITMAP *dbuf[2], *bbuf;
	int dbufptr = 0;
	int fcount = 0;
	TextureSt light, dark, weirdo;

	BITMAP *lightbmp, *darkbmp, *weirdobmp;
	PALETTE pal;

	allegro_init();
	set_window_close_hook(CloseProc);
	request_refresh_rate(100);
	text_mode(-1);

	install_keyboard();

	set_color_depth(SCR_BPP);
#ifdef _DEBUG
	set_gfx_mode(GFX_AUTODETECT_WINDOWED, SCR_W, SCR_H, 0, 0);
#else
	if(set_gfx_mode(GFX_AUTODETECT, SCR_W, SCR_H, 0, 0))
		set_gfx_mode(GFX_AUTODETECT_WINDOWED, SCR_W, SCR_H, 0, 0);
#endif
	AllocateTables();

	weirdobmp = load_pcx("weirdo.pcx", pal);
	lightbmp = load_pcx("brick.pcx", pal);
	darkbmp = load_pcx("brickd.pcx", pal);
	set_palette(pal);
	SetupTexture(&light, lightbmp);
	SetupTexture(&dark, darkbmp);
	SetupTexture(&weirdo, weirdobmp);
	destroy_bitmap(lightbmp);
	destroy_bitmap(darkbmp);
	destroy_bitmap(weirdobmp);

	/* decide what sort of buffering to use */
	dbuf[0] = create_video_bitmap(SCREEN_W, SCREEN_H);
	dbuf[1] = create_video_bitmap(SCREEN_W, SCREEN_H);

	if(!dbuf[1])
	{
		show_video_bitmap(dbuf[0]);
		UseVideoMemory = false;
	}
	else
		UseVideoMemory = true;

	bbuf = create_system_bitmap(SCREEN_W, SCREEN_H);

	fixed Corners[] =
	{
		ftofix(4), ftofix(4),
		ftofix(-4), ftofix(4),
		ftofix(-5), ftofix(2),
		ftofix(-5), ftofix(-2),
		ftofix(-4), ftofix(-4),
		ftofix(4), ftofix(-4),

		// 12
		ftofix(-8), ftofix(-10),
		ftofix(-8), ftofix(10),
		ftofix(-7), ftofix(10),

		// 18
		ftofix(-8), ftofix(20),
		ftofix(-7), ftofix(19),

		// 22
		ftofix(10), ftofix(23),
		ftofix(9), ftofix(22),

		// 26
		ftofix(10), ftofix(-15),
		ftofix(9), ftofix(-15),

		//30
		ftofix(30), ftofix(-50),
		ftofix(-11), ftofix(-50)
	};
	Wall TestWallA[6], TestWallB[5], TestWallC[4], TestWallD[4], TestWallE[4], TestWallF[4];
	Sector SectorA =
	{
		TestWallA, 6
	};
	Sector SectorB =
	{
		TestWallB, 5
	};
	Sector SectorC =
	{
		TestWallC, 4
	};
	Sector SectorD =
	{
		TestWallD, 4
	};
	Sector SectorE =
	{
		TestWallE, 4
	};
	Sector SectorF =
	{
		TestWallF, 4
	};

	SetupWall(&TestWallA[0], &light, &Corners[0], &Corners[2], NULL);
	SetupWall(&TestWallA[1], &dark, &Corners[2], &Corners[4], NULL);
	SetupWall(&TestWallA[2], &light, &Corners[4], &Corners[6], &SectorB);
	SetupWall(&TestWallA[3], &dark, &Corners[6], &Corners[8], NULL);
	SetupWall(&TestWallA[4], &light, &Corners[8], &Corners[10], NULL);
	SetupWall(&TestWallA[5], &dark, &Corners[10], &Corners[0], NULL);
	SetupSector(&SectorA);

	SetupWall(&TestWallB[0], &light, &Corners[6], &Corners[4], &SectorA);
	SetupWall(&TestWallB[1], &dark, &Corners[4], &Corners[16], NULL);
	SetupWall(&TestWallB[2], &dark, &Corners[16], &Corners[14], &SectorC);
	SetupWall(&TestWallB[3], &dark, &Corners[14], &Corners[12], NULL);
	SetupWall(&TestWallB[4], &dark, &Corners[12], &Corners[6], NULL);
	SetupSector(&SectorB);

	SetupWall(&TestWallC[0], &light, &Corners[18], &Corners[14], NULL);
	SetupWall(&TestWallC[1], &dark, &Corners[20], &Corners[18], &SectorD);
	SetupWall(&TestWallC[2], &light, &Corners[16], &Corners[20], NULL);
	SetupWall(&TestWallC[3], &dark, &Corners[14], &Corners[16], &SectorB);
	SetupSector(&SectorC);

	SetupWall(&TestWallD[0], &dark, &Corners[18], &Corners[20], &SectorC);
	SetupWall(&TestWallD[1], &dark, &Corners[20], &Corners[24], NULL);
	SetupWall(&TestWallD[2], &dark, &Corners[24], &Corners[22], &SectorE);
	SetupWall(&TestWallD[3], &dark, &Corners[22], &Corners[18], NULL);
	SetupSector(&SectorD);

	SetupWall(&TestWallE[0], &light, &Corners[22], &Corners[24], &SectorD);
	SetupWall(&TestWallE[1], &light, &Corners[24], &Corners[28], NULL);
	SetupWall(&TestWallE[2], &light, &Corners[28], &Corners[26], &SectorF);
	SetupWall(&TestWallE[3], &light, &Corners[26], &Corners[22], NULL);
	SetupSector(&SectorE);

	SetupWall(&TestWallF[0], &light, &Corners[26], &Corners[28], &SectorE);
	SetupWall(&TestWallF[1], &weirdo, &Corners[28], &Corners[32], NULL);
	SetupWall(&TestWallF[2], &weirdo, &Corners[32], &Corners[30], NULL);
	SetupWall(&TestWallF[3], &weirdo, &Corners[30], &Corners[26], NULL);
	SetupSector(&SectorF);

	unsigned int ang = 0;
	fixed pos[2] = {0, 0};
	Sector *CSector = &SectorA;

	time_t oldtime, newtime;
	oldtime = time(&oldtime);
	float fps;

	while(!key[KEY_ESC] && !quit)
	{
		rectfill(bbuf, 0, 0, SCREEN_W, SCREEN_H >> 1, 36);
		rectfill(bbuf, 0, SCREEN_H >> 1, SCREEN_W, SCREEN_H, 2);

		if(key[KEY_ALT] | key[KEY_ALTGR])
		{
			if(key[KEY_RIGHT])	CSector = AttemptMotion(CSector, &pos[0], &pos[1], -(LookupCos(ang) >> 3), LookupSin(ang) >> 3);
			if(key[KEY_LEFT])	CSector = AttemptMotion(CSector, &pos[0], &pos[1], LookupCos(ang) >> 3, -(LookupSin(ang) >> 3));
		}
		else
		{
			if(key[KEY_LSHIFT] | key[KEY_RSHIFT])
			{
				if(key[KEY_LEFT])	ang=(ang+64)&CIRCLE_MASK;
				if(key[KEY_RIGHT])	ang=(ang-64)&CIRCLE_MASK;
			}
			else
			{
				if(key[KEY_LEFT])	ang=(ang+32)&CIRCLE_MASK;
				if(key[KEY_RIGHT])	ang=(ang-32)&CIRCLE_MASK;
			}
		}

		if(key[KEY_LSHIFT] | key[KEY_RSHIFT])
		{
			if(key[KEY_UP])		CSector = AttemptMotion(CSector, &pos[0], &pos[1], LookupSin(ang) >> 2, LookupCos(ang) >> 2);
			if(key[KEY_DOWN])	CSector = AttemptMotion(CSector, &pos[0], &pos[1], -(LookupSin(ang) >> 2), -(LookupCos(ang) >> 2));
		}
		else
		{
			if(key[KEY_UP])		CSector = AttemptMotion(CSector, &pos[0], &pos[1], LookupSin(ang) >> 3, LookupCos(ang) >> 3);
			if(key[KEY_DOWN])	CSector = AttemptMotion(CSector, &pos[0], &pos[1], -(LookupSin(ang) >> 3), -(LookupCos(ang) >> 3));
		}

		sc = 1;
		BeginFrame();
		DrawSector(CSector, ang, pos[0], pos[1], 0, SCREEN_W-1);
		EndFrame(bbuf);
		fcount++;

		newtime = time(&newtime);

		if(newtime != oldtime)
		{
			fps = (float)fcount / (newtime - oldtime);
			fcount = 0;
			oldtime = newtime;
		}
		textprintf(bbuf, font, 0, 0, makecol(255,255,255), "Sectors in display: %d", sc);
		textprintf(bbuf, font, 0, 8, makecol(255,255,255), "FPS: %0.2f with refresh rate %dhz", fps, get_refresh_rate());

		if(UseVideoMemory)
			textprintf(bbuf, font, 0, 16, makecol(255,255,255), "Using page flipping");
		else
			textprintf(bbuf, font, 0, 16, makecol(255,255,255), "Using double buffering");

		if(UseVideoMemory)
		{
			blit(bbuf, dbuf[dbufptr], 0, 0, 0, 0, SCREEN_W, SCREEN_H);
			show_video_bitmap(dbuf[dbufptr]);
			dbufptr ^= 1;
		}
		else
		{
			vsync();
			blit(bbuf, dbuf[dbufptr], 0, 0, 0, 0, SCREEN_W, SCREEN_H);
		}
	}

	destroy_bitmap(dbuf[0]);
	destroy_bitmap(dbuf[1]);
	FreeTexture(&light);
	FreeTexture(&dark);
	FreeTexture(&weirdo);
	FreeTables();

	return 0;
}

END_OF_MAIN()